//
//  ES1Renderer.m
//  hw2_graphics
//
//  Created by slork on 1/12/10.
//  Copyright Stanford University 2010. All rights reserved.
//

#import "ES1Renderer.h"
#import "mo_audio.h"
#import "mo_touch.h"
#import "mo_accel.h"
#import "mo_gfx.h"
#import "Controls.h"
#import <vector>
#import <math.h>

using namespace std;

#define SRATE 24000
#define FRAMESIZE 256
#define NUMCHANNELS 2

// global
SAMPLE g_length = 1;     // distance between two pegs
float g_angle = 0;       // in radians, theta between two pegs
enum launchingStatus { INIT, LAUNCHING, LAUNCHED }; 
int g_launchingStatus = INIT; 

float g_min_freq = 100.0; 
float g_freq; 
float g_amplitude = 0.90;
// float g_ballradius = 0.1; 
float g_xAccel = 0.0;
float g_yAccel = 0.0; 

// bool g_noWallHit = true; 

vector<Keyloc> g_v3;    // first two points are always the slingshot; the third will become ball
vector<Ball> g_balls; 
GLuint g_texture[1];    // peg & ball

// buffer
SAMPLE g_vertices[FRAMESIZE*2];


// Implement audio callback here

 void audioCallback( Float32 * buffer, UInt32 numFrames, void* userData)
 {
	 SAMPLE x = 0;
	 SAMPLE inc = g_length / numFrames;

	 AudioData *data = (AudioData*) userData;
 
	 for (int i=0; i<numFrames; i++)
	 {

		buffer[2*i] = buffer[2*i+1] = data->sk->tick() + data->sk_wall->tick() + data->sk_ball->tick();  

		 
		 // x coordinate
		 g_vertices[2*i] = x; x += inc;
		 // y coordinate
		 g_vertices[2*i+1] = 0.3* data->sk->tick(); // buffer[2*i];
		 
		 
	 }
 }

void accelCallback( double x, double y, double z, void *data)
{
	Controls* c = Controls::instance();
	
	if (fabs(x) + fabs(y)> 3.5) {
		if (c->getClear()) g_balls.clear();
	}
	
	else {
	
		g_xAccel = x * c->getGravity();
		g_yAccel = -1*y * c->getGravity(); 
	}
}


void testLaunch(/*AudioData audioData*/)
{
	assert(g_v3.size()>=3);
	
	float x0 = g_v3[0].x;
	float y0 = g_v3[0].y;
	float x1 = g_v3[1].x;
	float y1 = g_v3[1].y;
	float x2 = g_v3[2].x;
	float y2 = g_v3[2].y; 
	
	float expected = (y1-y0)/(x1-x0)*(x2-x0);
	if ( (y2-y0 < expected + 0.5 ) && (y2-y0 > expected - 0.5 )  ) {
		g_launchingStatus = LAUNCHING; 
	}
}



bool testWallCollision (int i, AudioData audioData)
{	
	Controls* c = Controls::instance();
	float ballradius = c->getBall();
	float elasticityFactor = 1- (c->getElasticity());

	bool collided = false;
	
	if (g_balls[i].pos.getX() < (-1+ballradius)) {
		g_balls[i].vel.setX(g_balls[i].vel.getX() *-1.0 * elasticityFactor) ; 
		g_balls[i].pos.setX(-1+ballradius);
		collided = true; 
	}
	if (g_balls[i].pos.getX() > (1-ballradius)) {
		g_balls[i].vel.setX(g_balls[i].vel.getX() *-1.0  * elasticityFactor) ; 
		g_balls[i].pos.setX(1-ballradius);
		collided = true; 
	}
	if (g_balls[i].pos.getY() < (-1.5+ballradius)) {
		g_balls[i].vel.setY(g_balls[i].vel.getY() *-1.0  * elasticityFactor) ; 
		g_balls[i].pos.setY(-1.5+ballradius);
		collided = true; 
	}
	if (g_balls[i].pos.getY() > (1.5-ballradius)) {
		g_balls[i].vel.setY(g_balls[i].vel.getY() *-1.0  * elasticityFactor) ; 
		g_balls[i].pos.setY(1.5-ballradius);
		collided = true; 
	}
	if (collided) {
		audioData.sk_wall->setFrequency(g_balls[i].freqID *0.25);
		audioData.sk_wall->pluck(g_balls[i].vel.getLength()*0.5);
	}
	
	return collided; 
}

void testStringPassing( int i, AudioData audioData)
{
	float x = g_balls[i].pos.getX();
	float y = g_balls[i].pos.getY();
	float vel = g_balls[i].vel.getLength(); 
	
	
	assert(g_v3.size()>=2);
	
	float x0 = g_v3[0].x;
	float y0 = g_v3[0].y;
	float x1 = g_v3[1].x;
	float y1 = g_v3[1].y;
	
	float minX = min(x0, x1);
	float maxX = max(x0, x1);
	float minY = min(y0, y1);
	float maxY = max(y0, y1);
	
	
	float expected = (y1-y0)/(x1-x0)*(x-x0);
	if ( (y-y0 < expected + 0.1 ) && (y-y0 > expected - 0.1 ) && minX<x && x<maxX && minY<y && y<maxY ) 
	{
		audioData.sk->pluck(vel);
	}
	
}

bool testBallCollision(int i, AudioData audioData)
{
	bool ballCollision = false; 
	
	Controls* c = Controls::instance();
	float elasticityFactor = 1- (c->getElasticity()) * 0.2;
	
	Vector2D pos = g_balls[i].pos; 
	float diameter = 2 * (c->getBall());
	
	for (int j=0/*i+1*/; j<g_balls.size(); j++)
	{
		if (j!=i) {
		Vector2D pos2 = g_balls[j].pos; 
		Vector2D diff = pos-pos2; 
		
		if (diff.getLength() < diameter) 
		{
			// calculate new velocities assuming equal mass
			// NSLog(@"two balls collided!");
			Vector2D temp = g_balls[i].vel * elasticityFactor; 
			g_balls[i].vel = g_balls[j].vel * elasticityFactor;
			g_balls[j].vel = temp; 
			
			// make sure the balls don't overlap
			/*
			while (diff.getLength() < diameter) {
				pos = pos + diff*-1.0; //. g_balls[i].vel; 
			}
			 */

			audioData.sk_ball->setFrequency(g_balls[i].freqID + g_balls[j].freqID);
			audioData.sk_ball->pluck( (g_balls[i].vel.getLength()+ g_balls[j].vel.getLength())*0.5);
			ballCollision = true; 
		}
		}

	}
	return ballCollision;
	
}


// creates a new Ball and store in g_balls vector
void createNewBall()
{
	float x0 = g_v3[0].x;
	float y0 = g_v3[0].y;
	float x1 = g_v3[1].x;
	float y1 = g_v3[1].y;
	float x2 = g_v3[2].x;
	float y2 = g_v3[2].y; 	
	
	Vector2D s = Vector2D(x0-x1, y0-y1);  // vector from p1 to p0 (between two pegs)
	Vector2D r = Vector2D(x2-x1, y2-y1);  // vector from p1 to p2 (between one of the pegs and touch)
	
	float projMag = (s.dot(r) * (1/ pow(s.getLength(), 2)));
	Vector2D proj = s * projMag;
	
	Ball newball;
	newball.pos = Vector2D(x2, y2);
	newball.vel = proj - r; 
	
	// NSLog(@"ball's velocity upon creation= %f", newball.vel.getLength()); 
	newball.freqID = g_freq; 
	newball.alpha = 1.0; 
	
	// newball.brassOn = false; 
	
	g_balls.push_back(newball);
}


void touchCallback( NSSet* touches, UIView * view, const std::map<int, MoTouchTrack> & touchPts, void* data)
{
	AudioData* audioData = (AudioData*) data;
	
	int count = touchPts.size();
	
	// safety check, to prevent 1 peg being "stuck on", which occasionally happens
	if (count ==0) {g_v3.clear();}
	else if (count == 1) {
		for (UITouch *touch in touches) {
			if (touch.phase == UITouchPhaseEnded) {
				g_v3.clear(); 
				break;
			}
		}
	}
	
	// Update the vector
	for(UITouch * touch in touches){
		
		int k = (long) touch;
		CGPoint location = [touch locationInView:view];
		
		// Record the touch
		if( touch.phase == UITouchPhaseBegan ){
			//try to find the touch
			bool found = false;
			
			for(int i=0; i < g_v3.size(); i++){
				if(g_v3[i].key == k){
					found = true;
					break;
				}
			}
			if(!found)
			{
				Keyloc elem;
				elem.key = k;
				elem.x = (location.x)/160.-1.;
				elem.y = (location.y+7)/160.-1.5; // was /240-1. removed -
				g_v3.push_back(elem);
			}
		}
		else if( touch.phase == UITouchPhaseEnded || touch.phase == UITouchPhaseCancelled ){
			// Erase from the vector
			for(int i=0; i < g_v3.size(); i++){
				
				if(g_v3[i].key == k){
					g_v3.erase(g_v3.begin()+i);
					break;
				}
			}
		}
		
		else if (touch.phase == UITouchPhaseMoved) {
			for (int i=0; i<g_v3.size(); i++){
				if(g_v3[i].key == k) {
					g_v3[i].x = (location.x)/160.-1.;
					g_v3[i].y = (location.y+7)/160.-1.5; // was /240-1. removed -

					break;
				}
			}
		}
	}
	
	if( g_v3.size()==3 && g_launchingStatus != LAUNCHING) testLaunch(/**audioData*/);
	if (g_v3.size()<3 && g_launchingStatus == LAUNCHING) {
		createNewBall();
		audioData->sk->pluck(g_balls[g_balls.size()-1].vel.getLength());  // amplitude of the pluck is the velocity (i.e. amount of pull on rubberband)
		g_launchingStatus=LAUNCHED;
	}
	
	// uncomment below to have straight line with new slingshot creation despite sound that's still decaying
	/* 
	if (g_v3.size()==0 & g_launchingStatus == LAUNCHED) {
		g_launchingStatus=INIT;
	}
	 */

}

void render_texture_i(int index, int scale, const GLfloat img_vertices[])
{
	int i = index;
	
    static const GLfloat normals[] = {
        0, 0, 1,
        0, 0, 1,
        0, 0, 1,
        0, 0, 1
    };
	
	static const GLfloat texCoords[] = {
        0, 1,
        1, 1,
        0, 0,
        1, 0
    };
	
	
	glPushMatrix();
	glTranslatef(g_v3[i].x, g_v3[i].y, 0);
	glScalef(scale, scale, scale); 
	
	// vertex
	glVertexPointer( 2, GL_FLOAT, 0, img_vertices );
	glEnableClientState(GL_VERTEX_ARRAY );
	
	// color
	if (i==0 || i==1) {	
		// NSLog(@"rendering peg #%d", i); 
		glColor4f( 1.0, 0, 0, 1.0 );
	} // pegs
	else glColor4f(0.8, 0, 0.8, 1.0); // other touched fingers
	
	// normal
	glNormalPointer( GL_FLOAT, 0, normals );
	glEnableClientState( GL_NORMAL_ARRAY );
	
	// texture coordinate
	glTexCoordPointer( 2, GL_FLOAT, 0, texCoords );
	glEnableClientState( GL_TEXTURE_COORD_ARRAY );
	
	// triangle strip
	glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );
	
	glPopMatrix();	
}


void drawPegs()
{
	// glEnable( GL_TEXTURE_2D ); // IMPORTANT
	
	// NSLog(@"inside drawPegs!!!");

	static const GLfloat pegVertices[] = {
        -0.2f,  -0.2f,
        0.2f,  -0.2f,
        -0.2f,   0.2f,
        0.2f,   0.2f,
    };
	
	int scale = 1; 
	
	
	int numPegs = min(2, (int)g_v3.size());
	// NSLog(@"in drawPegs, numPegs = %d", numPegs);
	
	for (int i=0; i<numPegs; i++) {
		render_texture_i(i, scale, pegVertices);
	}
}	



void drawString(struct AudioData audioData)
{
	// glDisable( GL_TEXTURE_2D ); // IMPORTANT
	
	float x0 = g_v3[0].x;
	float y0 = g_v3[0].y;
	float x1 = g_v3[1].x;
	float y1 = g_v3[1].y;
	float x2 = g_v3[2].x;
	float y2 = g_v3[2].y; 

	// straight line, no sound
	if (g_launchingStatus == INIT) {
		
		// draw straight line from p0 to p1
		GLfloat launchLineVertices[4] = {x0, y0, x1, y1};
		
		glPushMatrix();
		
		glVertexPointer(2, GL_FLOAT, 0, launchLineVertices);
		glEnableClientState(GL_VERTEX_ARRAY);
		
		glColor4f( 1.0f, 0.0f, 0.0f, 1.0f );
		glDrawArrays( GL_LINE_STRIP, 0, 2 ); 		
		
	}
				
	// third touch intersected the string: ready to launch!
	else if (g_launchingStatus == LAUNCHING) {
		
		// draw line segment from p0 to p2 to p1
		GLfloat launchLineVertices[6] = {x0, y0, x2, y2, x1, y1};

		
		glPushMatrix();
		
		glVertexPointer(2, GL_FLOAT, 0, launchLineVertices);
		glEnableClientState(GL_VERTEX_ARRAY);
		
		glColor4f( 0.0f, 0.0f, 1.0f, 1.0f );
		glDrawArrays( GL_LINE_STRIP, 0, 3 ); 
	}
		
	// straight line sound wave-form
	else if (g_launchingStatus == LAUNCHED) {
	
		bool flip = false;
		if (x0>x1) flip = true;
	
		float x_dif = x1-x0; 
		float y_dif = y1-y0; 
	
		g_length = sqrt(pow((x_dif),2) + pow((y_dif),2));
		// NSLog(@"g_length = %f", g_length); 
		
		g_angle = atan(y_dif / x_dif) * 180.0 / M_PI; 
		float rotate = g_angle;
		if (flip) rotate += 180; 
	
		glPushMatrix();
	
		glTranslatef( x0, y0, 0 );
		glRotatef(rotate, 0, 0, 1);

		glVertexPointer(2, GL_FLOAT, 0, g_vertices);
		glEnableClientState(GL_VERTEX_ARRAY);
	
		glColor4f( 1.0f, 0.0f, 0.0f, 1.0f );
		glDrawArrays( GL_LINE_STRIP, 0, FRAMESIZE );
	
		glPopMatrix(); 
		
		g_freq = g_min_freq * pow(10, 1/g_length); 
		audioData.sk->setFrequency(g_freq); // SET FREQUENCY TO BE DISTANCE BETWEEN PEGS
	
	}
}

void drawOtherTouches()
{
	Controls* c = Controls::instance();
	float scale = c->getBall() * 10.0;
	
	static const GLfloat otherTouchesVertices[] = {
		-0.1,  -0.1,
		0.1,  -0.1,
		-0.1,   0.1,
		0.1,    0.1
	};
	
	// for now, only one extra touch allowed... 
	/*
	for (int i=2; i< g_v3.size(); i++) {
		render_texture_i(i, otherTouchesVertices);
	}
	 */
	render_texture_i(2, scale, otherTouchesVertices);
}

void drawBalls()
{
	// glEnable( GL_TEXTURE_2D ); // IMPORTANT
	Controls* c = Controls::instance();
	float ballscale = c->getBall() *10.0;
	
	static const GLfloat squareVertices[] = {
		-0.1,  -0.1,
		0.1,  -0.1,
		-0.1,   0.1,
		0.1,    0.1
    };
	
    static const GLfloat normals[] = {
        0, 0, 1,
        0, 0, 1,
        0, 0, 1,
        0, 0, 1
    };
    
    static const GLfloat texCoords[] = {
        0, 1,
        1, 1,
        0, 0,
        1, 0
    };
	
    // clear
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);
    
    // for each entity
    for( int i = 0; i < g_balls.size(); i++ )
    {
        glPushMatrix();
		
        // translate
        glTranslatef( g_balls[i].pos.getX(), g_balls[i].pos.getY(), 0 );
		
		glScalef(ballscale, ballscale, ballscale);
        
        // vertex
        glVertexPointer( 2, GL_FLOAT, 0, squareVertices );
        glEnableClientState(GL_VERTEX_ARRAY );
		
        // color
        glColor4f( 0.8, 0, 0.8, g_balls[i].alpha);
        
        // normal
        glNormalPointer( GL_FLOAT, 0, normals );
        glEnableClientState( GL_NORMAL_ARRAY );
        
        // texture coordinate
        glTexCoordPointer( 2, GL_FLOAT, 0, texCoords );
        glEnableClientState( GL_TEXTURE_COORD_ARRAY );
        
        // triangle strip
        glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );
        
        glPopMatrix();
    }
}


void drawAll(struct AudioData audioData)
{
	glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
    glClear( GL_COLOR_BUFFER_BIT );       
	
	// why is it that if I don't draw balls first, the other things don't get rendered once the ball has been created? 
	if(g_balls.size()>0) {
		drawBalls();
	}
	
	if (g_v3.size()>0) {
		drawPegs();
	}
		
	if(g_v3.size()>1) {
		glDisable( GL_TEXTURE_2D ); // IMPORTANT
		drawString(audioData);
		glEnable( GL_TEXTURE_2D ); // IMPORTANT

	}
	
	if(g_v3.size()>2) {
		drawOtherTouches();
	}
	

}	

void updateBall(AudioData audioData)
{
	for (int i=0; i<g_balls.size(); i++) 
	{
		Vector2D accel = Vector2D(g_xAccel, g_yAccel);
		
		// velocity reduces with time (~air resistance)
		Controls* c = Controls::instance();
		g_balls[i].vel *= 1- 0.01*c->getDamping();
		
		// automatically fades and erases the ball if the velocity becomes ~0. 
		/*
		 float velMagnitude  = g_balls[i].vel.getLength();
		 if ( velMagnitude < 0.01) g_balls[i].alpha = velMagnitude*100; 
		 if ( velMagnitude > 0.01) g_balls[i].alpha = 1.0; 
		 if (g_balls[i].alpha < 0.1) g_balls.erase(g_balls.begin()+i); 
		 */
		
		// v = da/dt
		g_balls[i].vel = g_balls[i].vel + accel; 
		
		Vector2D oldPos = g_balls[i].pos; 
		
		// x = dv/dt
		g_balls[i].pos = g_balls[i].pos + g_balls[i].vel*0.05;

		
		
		// collision handling on walls
		testWallCollision(i, audioData); 
		
		// collision handling with the string
		if (g_v3.size()>1) {
			testStringPassing(i, audioData); 
		}
		
		// collision handling with other balls
		if (testBallCollision(i, audioData)) g_balls[i].pos = oldPos; 
		
	}
}


@implementation ES1Renderer

// Create an ES 1.1 context
- (id) init
{
	if (self = [super init])
	{
		context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];
        
        if (!context || ![EAGLContext setCurrentContext:context])
		{
            [self release];
            return nil;
        }
		
		// Create default framebuffer object. The backing will be allocated for the current layer in -resizeFromLayer
		glGenFramebuffersOES(1, &defaultFramebuffer);
		glGenRenderbuffersOES(1, &colorRenderbuffer);
		glBindFramebufferOES(GL_FRAMEBUFFER_OES, defaultFramebuffer);
		glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
		glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_RENDERBUFFER_OES, colorRenderbuffer);
	}
	
	
	// enable texture mapping
    glEnable( GL_TEXTURE_2D );
    // enable blending
    glEnable( GL_BLEND );
    // blend function
    // glBlendFunc( GL_ONE, GL_ZERO );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA ); // this is better
	
    // generate texture name
    glGenTextures( 1, &g_texture[0] );
	
    // bind the texture
    glBindTexture( GL_TEXTURE_2D, g_texture[0] );
    // setting parameters
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    // load the texture
    MoGfx::loadTexture( @"ball", @"png" );
		
	NSLog(@"clearing g_V3 vector in the init fn");
	
	// init audio
	
	// setup the STK itself
	Stk::setSampleRate( SRATE );
	Stk::showWarnings (false);
	
	audioData.sk = new StifKarp(g_min_freq);
	audioData.sk_wall = new StifKarp(g_min_freq);
	audioData.sk_ball = new StifKarp(g_min_freq);

	
	// audio
	NSLog(@"Initializing Audio");
	bool result = MoAudio::init(SRATE, FRAMESIZE, NUMCHANNELS);
	
	if (!result)
	{
		NSLog(@"cannot initialize real-time audio!");
		return self;
	}
	
	// start the audio layer, registering a callback method
	result = MoAudio::start( audioCallback, &audioData);
	if (!result)
	{
		NSLog(@"cannot start real-time audio!");
		return self;
	}
	
	// touch
	MoTouch::addCallback( touchCallback, &audioData);
	
	// accel
	MoAccel::addCallback( accelCallback, NULL);
	
	return self;
}

- (void) render
{
	
	Controls* c = Controls::instance();
	if(c->getClearTouches()) {
		g_v3.clear();
		c->setClearTouches(false);
	}
	if(c->getClearBalls()) {
		g_balls.clear();
		c->setClearBalls(false);
	}
	
    // Replace the implementation of this method to do your own custom drawing
    
	
	// This application only creates a single context which is already set current at this point.
	// This call is redundant, but needed if dealing with multiple contexts.
    [EAGLContext setCurrentContext:context];
    
	// This application only creates a single default framebuffer which is already bound at this point.
	// This call is redundant, but needed if dealing with multiple framebuffers.
    glBindFramebufferOES(GL_FRAMEBUFFER_OES, defaultFramebuffer);
    glViewport(0, 0, backingWidth, backingHeight);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	glOrthof( -1.0f, 1.0f, 1.5f, -1.5f, -1.0f, 1.0f ); 
    glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	updateBall(audioData);
	drawAll(audioData);
		
    // glColorPointer(4, GL_UNSIGNED_BYTE, 0, squareColors);
    // glEnableClientState(GL_COLOR_ARRAY);
    
    // glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    
	// This application only creates a single color renderbuffer which is already bound at this point.
	// This call is redundant, but needed if dealing with multiple renderbuffers.
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
    [context presentRenderbuffer:GL_RENDERBUFFER_OES];
}

- (BOOL) resizeFromLayer:(CAEAGLLayer *)layer
{	
	// Allocate color buffer backing based on the current layer size
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
    [context renderbufferStorage:GL_RENDERBUFFER_OES fromDrawable:layer];
	glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_WIDTH_OES, &backingWidth);
    glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_HEIGHT_OES, &backingHeight);
	
    if (glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES) != GL_FRAMEBUFFER_COMPLETE_OES)
	{
		NSLog(@"Failed to make complete framebuffer object %x", glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES));
        return NO;
    }
    
    return YES;
}

- (void) dealloc
{
	// Tear down GL
	if (defaultFramebuffer)
	{
		glDeleteFramebuffersOES(1, &defaultFramebuffer);
		defaultFramebuffer = 0;
	}

	if (colorRenderbuffer)
	{
		glDeleteRenderbuffersOES(1, &colorRenderbuffer);
		colorRenderbuffer = 0;
	}
	
	// Tear down context
	if ([EAGLContext currentContext] == context)
        [EAGLContext setCurrentContext:nil];
	
	[context release];
	context = nil;
	
	[super dealloc];
}

@end
