/*
 *  Vector2D.h
 *  oh_hw2
 *
 *  Created by Jieun Oh on 1/31/10.
 *  Copyright 2010 Stanford University. All rights reserved.
 *
 */

#ifndef VECTOR2D_H
#define VECTOR2D_H

#include <cmath>

	
	/***************************************************/
	/* taken from STK Vector3D class. added some more functionalities
	/***************************************************/
	
	class Vector2D 
		{
			
		public:
			//! Default constructor taking optional initial X, Y values.
			Vector2D( float x = 0.0, float y = 0.0 ) { setXY( x, y ); };
			
			Vector2D( const Vector2D & rhs) {*this = rhs;};

			
			//! Get the current X value.
			float getX( void ) { return X_; };
			//! Get the current Y value.
			float getY( void ) { return Y_; };
			//! Calculate the vector length.
			float getLength( void );
			// Calculate dot product.
			float dot( Vector2D v1);
			
			
			void operator= (const Vector2D & rhs);
			
			Vector2D operator+ (const Vector2D & rhs);
			Vector2D operator- (const Vector2D & rhs);
			Vector2D operator* (float scalar);
			
			void operator+= (const Vector2D & rhs);
			void operator-= (const Vector2D & rhs);
			void operator*= (float scalar); 
			
			//! Set the X, Y values simultaniously.
			void setXY( float x, float y ) { X_ = x; Y_ = y;};
			//! Set the X value.
			void setX( float x ) { X_ = x; };
			//! Set the Y value.
			void setY( float y ) { Y_ = y; };

		protected:
			float X_;
			float Y_;
		};
	
	inline float Vector2D::getLength( void )
	{
		float temp;
		temp = X_ * X_;
		temp += Y_ * Y_;
		temp = sqrt( temp );
		return temp;
	}
	
	inline float Vector2D::dot( Vector2D v1)
	{
		return ( v1.getX()*X_ + v1.getY()*Y_ ); 
	}
	
	inline void Vector2D::operator= (const Vector2D & rhs)
	{
		X_=rhs.X_; 
		Y_=rhs.Y_;	
	}
	
	inline Vector2D Vector2D::operator +(const Vector2D & rhs) 
	{ 
		Vector2D result = *this; 
		result += rhs; 
		return result; 
	}
	
	inline Vector2D Vector2D::operator -(const Vector2D & rhs) 
	{ 
		Vector2D result = *this; 
		result -= rhs; 
		return result; 
	}
	
	inline Vector2D Vector2D::operator *(float scalar) 
	{ 
		Vector2D result = *this; 
		result *= scalar; 
		return result; 
	}
	
	inline void Vector2D::operator +=(const Vector2D & rhs)
	{ 
		X_ += rhs.X_; 
		Y_ += rhs.Y_;
	}
	inline void Vector2D::operator -=(const Vector2D & rhs)
	{ 
		X_ -= rhs.X_; 
		Y_ -= rhs.Y_;
	}
	inline void Vector2D::operator *=(float scalar) 
	{ 
		X_ *= scalar; 
		Y_ *= scalar;
	}

	
#endif


