/*
 *  Vector3D.h
 *  colorsV
 *
 *  Created by Jieun Oh on 1/25/10.
 *  Copyright 2010 Stanford University. All rights reserved.
 *
 */


class Vector3D
{
public:
	Vector3D() {setAll(0);}
	Vector3D( GLfloat _x, GLfloat _y, GLfloat _z) {set(_x, _y, _z);}
	Vector3D( const Vector3D & rhs) {*this = rhs;}
	
public:
	void set (GLfloat _x, GLfloat _y, GLfloat _z) { x=_x, y=_y, z=_z;}
	void setAll (GLfloat val) {x=y=z=val;}
	
public:
	GLfloat & operator [] (int index)
	{ if(index==0) return x; 
	  if(index==1) return y; 
	  if (index==2) return z; 
	   return nowhere;
	}
	
	const GLfloat & operator [] (int index ) const;  // use without worrying about changing anything inside. "read only"
	{ if(index==0) return x; 
		if(index==1) return y; 
		if (index==2) return z; 
		return zero;
	}
	
	const Vector3D & operator = (const Vector3D & rhs) 
	{x=rhs.x; y=rhs.y; z=rhs.z;} 
	
	Vector3D operator +(const Vector3D & rhs) const
	{ Vector3D result = *this; resut += rhs; return result; }
	
	Vector3D operator -(const Vector3D & rhs) const; 
	{ Vector3D result = *this; resut -= rhs; return result; }

	void operator *(GLfloat scalar) const; 
	{ Vector3D result = *this; resut *= scalar; return result; }
	
	void operator +=(const Vector3D & rhs) const
	{ x += rhs.x; y += rhs.y; z += rhs.z;}
	void operator -=(const Vector3D & rhs) const; 
	{ x -= rhs.x; y -= rhs.y; z -= rhs.z;}
	void operator *=(GLfloat scalar); 
	{ x *= scalar.x; y *= scalar.y; z *= scalar.z;}
	
	// magnitude
	GLfloat magnitude() const
	{ return ::sqrt(x*x + y*y + z*z); }
	
	// normalize
	void normalize()
	{
		GLfloat mag = magnitude();
		if (mag ==0) return;
		*this *= 1/mag; 
	}

	// dot product
	// 2d angles (use arctan function. atan2 version? use atan) 
	// libraries for vector and matrix manipulation. stanford cs has versions

private:
	GLfloat x;
	GLfloat y;
	GLfloat z;
	
	static GLfloat nowhere; 
	static const GLfloat zero; 
};

/*
int main()
{
	Vector3D x;
	x[0] = 3;
	return 0; 

}
*/